/*
 * Licensed Materials - Property of IBM,
 *     Copyright IBM Corp. 2010, 2011  All Rights Reserved
 */

package com.ibm.tools.attach.javaSE;

import java.io.File;
import java.io.IOException;
import java.util.Timer;
import java.util.TimerTask;

class FileLock {
	long fileDescriptor;
	String lockFilepath;
	int fileMode;
	private boolean locked;
	private static FilelockTimer fileLockWatchdogTimer;
	static class syncObject {};
	private static syncObject shutdownSync = new syncObject();
	static boolean terminated;

	boolean isLocked() {
		return locked;
	}

	FileLock(String filePath, int mode) {
		locked = false;
		if (null == filePath) {
			throw new NullPointerException("filePath is null"); //$NON-NLS-1$
		}
		lockFilepath= filePath;
		fileMode = mode;
	}

	/**
	 * locks a file using the J9 port library functions.
	 * @param blocking if true, the function waits until the it obtains the file lock or times out.
	 * @return true is file lock obtained.
	 * @throws IOException if the file is already locked.
	 * @note locking and unlocking must be done by the same thread.
	 */
	boolean lockFile(boolean blocking) throws IOException {
		if (IPC.loggingEnabled ) {
			IPC.logMessage(blocking? "locking file ": "non-blocking locking file ", lockFilepath);  //$NON-NLS-1$//$NON-NLS-2$
		}

		final int FILE_LOCK_TIMEOUT = 60 * 1000; /* time in milliseconds */
		if (locked) {
			throw new IOException("file already locked"); //$NON-NLS-1$
		}

		/* lock will usually be uncontended, so don't start a watchdog unless necessary */
		fileDescriptor = lockFileImpl(lockFilepath, fileMode, false);
		locked = (0 <= fileDescriptor); /* negative values indicate error, non-negative values (including 0) are valid FDs */

		if (!locked && blocking) { /* try again, this time with a blocking lock and a timeout */
			FileLockWatchdogTask wdog = new FileLockWatchdogTask();
			IPC.logMessage("lock failed, trying blocking lock"); //$NON-NLS-1$
			synchronized (shutdownSync) { /* shutdown is called from a different thread */
				if (!terminated && (null == fileLockWatchdogTimer)) {
					fileLockWatchdogTimer = new FilelockTimer("file lock watchdog"); //$NON-NLS-1$
				}
				if (null != fileLockWatchdogTimer) { /* check if the VM is shutting down */
					/*
					 * The file lock timeout is to recover from pathological conditions such as a hung process which is holding the file lock.
					 * Under this condition, the process will break the lock.
					 * This timeout affects only the operation of the attach API.  It may delay the start of the attach API (but will not affect
					 * other aspects of the VM or application launch) or delay an attach attempt.  It will not delay the VM or attach API shutdown.
					 */
				fileLockWatchdogTimer.schedule(wdog, FILE_LOCK_TIMEOUT);
				}
				else {
					return false;
				}
			}
			fileDescriptor = lockFileImpl(lockFilepath, fileMode, true);
			synchronized (shutdownSync) {
				if (null != fileLockWatchdogTimer) {
					wdog.cancel();
				}
			}
			locked = (0 <= fileDescriptor);
		}

		return locked;
	}

	void unlockFile() {
		if (IPC.loggingEnabled ) {
			IPC.logMessage("unlocking file ", lockFilepath);  //$NON-NLS-1$
		}
		if (locked && (fileDescriptor >= 0)) {
			unlockFileImpl(fileDescriptor);
		}
		locked = false;
	}

	public static void shutDown() {
		/*
		 * After the last live reference to a Timer object goes away and all
		 * outstanding tasks have completed execution, the timer's task
		 * execution thread terminates
		 */
		synchronized (shutdownSync) {
			terminated = true;
			if (null != fileLockWatchdogTimer) {
				fileLockWatchdogTimer.cancel();
				fileLockWatchdogTimer.purge();
			}
			fileLockWatchdogTimer = null;
		}
	}

	private static native long lockFileImpl(String filePath, int mode, boolean blocking);

	private static native int unlockFileImpl(long fileDesc);

	class FileLockWatchdogTask extends TimerTask {
		@Override
		/**
		 * Wait for a timeout and delete and recreate the lock file.  Interrupt this thread if the lock is obtained before the timeout.
		 * @return true if the thread was interrupted or the lock file was replaced.
		 */
		public void run() {
			IPC.logMessage("waitAndCheckLock recreating ", lockFilepath); //$NON-NLS-1$
			/* recreate the file */
			File theFile = new File(lockFilepath);
			if (!theFile.renameTo(new File(theFile.getParent(), '.'
					+ "trash_" + IPC.getRandomNumber()))) { //$NON-NLS-1$
				IPC.logMessage("waitAndCheckLock could not rename ", theFile.getName()); //$NON-NLS-1$
				theFile.renameTo(new File(theFile.getParent(), '.' + "trash_" + IPC.getRandomNumber()));  //$NON-NLS-1$
				/* retry once.  If it fails more than once, it's a systemic problem */
			}
			unlockFile(); /* unlocks the file if this process, and closes the file descriptor to break the wait */
			/* delete the file if it's there */
			if (!theFile.delete()) {
				IPC.logMessage("waitAndCheckLock could not delete ", theFile.getAbsolutePath()); //$NON-NLS-1$
			} else {
				IPC.logMessage("waitAndCheckLock deleted ", theFile.getAbsolutePath()); //$NON-NLS-1$
			}
			IPC.logMessage("waitAndCheckLock normal return "); //$NON-NLS-1$
			return;
		}
	}
}
